-- parse_gen.gs
-- functions for parser generator


-- basic grammar functions

terminals :: Gram t n -> [t]
terminals (terms,_,_,_) = terms

nonterminals :: Gram t n -> [n]
nonterminals (_,nonterms,_,_) = nonterms

productions :: Gram t n -> [Production t n]
productions (_,_,prods,_) = prods

goal_symbol :: Gram t n -> n
goal_symbol (_,_,_,goal) = goal


-- fixed-point iteration

fixed_point :: Eq a => (a -> a) -> a -> a
fixed_point f x
	| y == x    = x
	| otherwise = fixed_point f y
			where y = f x

fixed_point2 :: Eq [a] => ([a] -> [a]) -> [a] -> [a] -> [a]
fixed_point2 f to_process processed
	| y == []  = processed
	| otherwise = fixed_point2 f y (y ++ processed)
			where
			  y = f to_process \\ processed

-- set processing functions

member :: Eq a => a -> [a] -> Bool
member x [] = False
member x (y : ys)
	| x == y    = True
	| otherwise = member x ys

union :: Eq a => [a] -> [a] -> [a]
union [] ys = ys
union (x:xs) ys
	| member x ys = union xs ys
	| otherwise  = x : union xs ys

extract :: Eq a => a -> [a] -> [a]
extract x [] = []
extract x (y:ys)
	| x == y    = extract x ys
	| otherwise = y : extract x ys

set_diff :: Eq a => [a] -> [a] -> [a]
set_diff xs [] = xs
set_diff xs (y:ys)
	| member y xs = set_diff (extract y xs) ys
	| otherwise   = set_diff xs ys


-- Association list functions

type Assoc a b = [(a,b)]

lookup :: Eq a => a -> Assoc a b -> b
lookup a ((a1,b1):abs)
	| a == a1   = b1
	| otherwise = lookup a abs


-- Nullable, First, Follow and Can_end

-- Normally the First sets can contain the empty symbol and the Follow
-- sets can contain the end symbol. Since the functions are
-- polymorphic over the terminals set, we can't do this.
-- Hence, Nullable x means that x can derive the empty string and
-- Can_end n means that the nonterminal n can be at the end of a
-- derivation

nullable_map :: Eq (Assoc n Bool) => Gram t n -> Assoc n Bool
nullable_map gram
	= fixed_point nullable1 [(n,False) | n <- nonterminals gram]
	where
	  nullable1 null_map
		= [(n,nullable_nont n (productions gram)) |
				n <- nonterminals gram]
		where
		  nullable_nont n [] = False
		  nullable_nont n ((m,rhs):prods)
			| n == m     = nullable rhs null_map ||
				       nullable_nont n prods
			| otherwise  = nullable_nont n prods

nullable :: Eq n => [Either t n] -> Assoc n Bool -> Bool
nullable [] 		 null_map = True
nullable (Left t : rhs)  null_map = False
nullable (Right n : rhs) null_map = lookup n null_map &&
				    nullable rhs null_map

normalize []    l' = []
normalize (a:l) l' | member a l' = a : normalize l l'
                   | otherwise   = normalize l l'

first_map :: Eq (Assoc n [t])  => Gram t n -> Assoc n Bool -> Assoc n [t]
first_map gram null_map
	= fixed_point first1 [(n,[]) | n <- nonterminals gram]
	where
	  first1 fir_map
		= [(n,normalize (terminals gram)
				(first_nont n (productions gram))) |
				n <- nonterminals gram]
		where
		  first_nont n []   = []
		  first_nont n ((m,rhs):prods)
			| n == m    = union (first rhs null_map fir_map)
					    (first_nont n prods)
			| otherwise = first_nont n prods

first :: (Eq n, Eq t) => [Either t n] -> Assoc n Bool -> Assoc n [t] -> [t]
first []              null_map fir_map = []
first (Left t : rhs)  null_map fir_map = [t]
first (Right n : rhs) null_map fir_map
	| lookup n null_map = union (lookup n fir_map) (first rhs null_map fir_map)
	| otherwise         = lookup n fir_map


follow_map :: Eq (Assoc n [t]) =>
		Gram t n -> Assoc n Bool -> Assoc n [t] -> Assoc n [t]
follow_map gram null_map fir_map
	= fixed_point follow1 [(n,[]) | n <- nonterminals gram]
	where
	  follow1 fol_map
		= [(n,normalize (terminals gram)
				(follow_nont n (productions gram))) |
				n <- nonterminals gram]
		where
		  follow_nont n [] = []
		  follow_nont n ((m,rhs):prods)
			= union (follow n rhs m null_map fir_map fol_map)
				(follow_nont n prods)

follow :: (Eq n, Eq t) =>
	n -> [Either t n] -> n -> Assoc n Bool -> Assoc n [t]
		-> Assoc n [t] -> [t]
follow n []              m null_map fir_map fol_map = []
follow n (Left t : rhs)  m null_map fir_map fol_map
	= follow n rhs m null_map fir_map fol_map
follow n (Right n1 : rhs) m null_map fir_map fol_map
	| n == n1 && (nullable rhs null_map)
		= union (lookup m fol_map)
			(union (first rhs null_map fir_map)
			       (follow n rhs m null_map fir_map fol_map))
	| n == n1   = union (first rhs null_map fir_map)
			    (follow n rhs m null_map fir_map fol_map)
	| otherwise = follow n rhs m null_map fir_map fol_map

can_end_map :: Eq (Assoc n Bool) =>
		Gram t n -> Assoc n Bool -> Assoc n Bool
can_end_map gram null_map
	= fixed_point end1 [(n,False) | n <- nonterminals gram]
	where
	  end1 end_map
		= [(n,end_nont n (productions gram)) |
				n <- nonterminals gram]
		where
		  end_nont n [] = n == goal_symbol gram
		  end_nont n ((m,rhs):prods)
			= lookup m end_map && can_end n rhs null_map
			  || end_nont n prods

can_end :: Eq n => n -> [Either t n] -> Assoc n Bool -> Bool
can_end n []              null_map = False
can_end n (Left t : rhs)  null_map = can_end n rhs null_map
can_end n (Right n1 : rhs) null_map
	| n == n1 = nullable rhs null_map || can_end n rhs null_map
	| otherwise = can_end n rhs null_map

fol_end gram
	= (fol_map,end_map)
	where
	  null_map = nullable_map gram
	  fir_map = first_map gram null_map
	  fol_map = follow_map gram null_map fir_map
	  end_map = can_end_map gram null_map


-- Functions on items, item sets and collections of item sets

closure :: (Eq n, Eq [Item t n]) => Gram t n -> [Item t n] -> [Item t n]
closure gram items
	= fixed_point2 closure1 items items
	where
	  closure1 items = foldr union items
			     [closure2 after | (n,before,after) <- items]
	  closure2 []                = []
	  closure2 (Left t : after)  = []
	  closure2 (Right n : after)
		= [(m,[],rhs) | (m,rhs) <- productions gram, m == n]

goto :: (Eq n, Eq [Item t n], Eq (Either t n)) =>
	 Gram t n -> [Item t n] -> Either t n ->[Item t n]
goto gram items symbol = snd (goto_kernel gram items symbol)
goto_kernel gram items symbol
	= (k,closure gram k)
        where k = [(n,s1:before,after) |
			(n,before,s1:after) <- items,
			s1 == symbol]

collection_of_item_sets :: (Eq (Either t n), Eq [([Item t n],[Item t n])]) =>
				Gram t n -> n -> [[Item t n]]
collection_of_item_sets gram start
	= collect goto1 initial
	where
          collect g i = col [] [i]
              where
                col seen [] = []
                col seen ((k,c):r) = if k `elem` seen
                                     then col seen r
                                     else c : col (k:seen) ((g c)++r)
	  initial = (i, closure gram i) where i = [(start,[],[goal])]
	  goal = Right (goal_symbol gram)
          symbols = [Right n | n <- nonterminals gram]
                    ++ [Left t | t <- terminals gram]
	  goto1 items = [goto_kernel gram items sym |
			   sym <- normalize symbols 
                                  [sym| (_,_,sym:_) <- items]]

slr_table gram
	= map make_row collection
	where
	  collection
		= collection_of_item_sets gram start \\ [[]]
	  start = goal_symbol gram ++ "\'"
	  (fol_map0,end_map0) = fol_end gram
	  fol_map = (start,[]) : fol_map0
	  end_map = (start,True) : end_map0
	  items_to_num = ([],0) : zip collection [1..]
	  prod_to_num = zip (productions gram) [1..]
	  make_row items = (lookup items items_to_num,
			    map (actions items) (terminals gram),
			    map item_to_prodnum
			    	(filter reduce_at_end items),
			    map ((\x -> lookup x items_to_num) .
				 goto gram items . Right)
				(nonterminals gram))
	  actions items t = (lookup (goto gram items (Left t))
	  			items_to_num,
			     map item_to_prodnum
				(filter (reduce_at t) items))
	  reduce_at t (n,_,[]) = member t (lookup n fol_map)
	  reduce_at t _        = False
	  reduce_at_end (n,_,[]) = lookup n end_map
	  reduce_at_end _        = False
	  item_to_prodnum (s,_,[])
		| s == start = 0
	  item_to_prodnum (n,before,[])
		| otherwise  = (lookup (n,(reverse before)) prod_to_num)

check_conflicts gram table
	= concat (map check_row table)
	where
	  check_row (state,actions,red_end,gotos)
		= concat (map (check_action state)
			      (zip (terminals gram) actions))
		  ++ check_red_end state red_end

	  check_action state (sym,(s,[]))  = ""
	  check_action state (sym,(0,[r])) = ""
	  check_action state (sym,(s,reds@[r]))
		| s /= 0 = "Shift/reduce conflict in state "
			   ++ show state ++ " at symbol "
			   ++ sym ++": "
			   ++ show_action (s,reds) ++"\n"
	  check_action state (sym,(0,reds@(r1:r2:_)))
		         = "Reduce/reduce conflict in state "
			   ++ show state ++ " at symbol "
			   ++ sym ++": "
			   ++ show_action (0,reds) ++"\n"
	  check_action state (sym,(s,reds@(r1:r2:_)))
		| s /= 0 = "Shift/reduce/reduce conflict in state "
			   ++ show state ++ " at symbol "
			   ++ sym ++": "
			   ++ show_action (s,reds) ++"\n"

	  check_red_end state [] = ""
	  check_red_end state [r] = ""
	  check_red_end state reds@(r1:r2:_)
		 = "Reduce/reduce conflict in state "
		   ++ show state ++ " at END: "
		   ++ concat (map show_red reds) ++"\n"

	  show_action (shift,reduce)
		= show_shift shift ++ concat (map show_red reduce)
	  show_shift 0 = ""
	  show_shift s  = "s"++ show s
	  show_red 0 = "acc"
	  show_red p = "r" ++ show p

parse_gen gram_file main_name scan_name parser_file
  = readFile gram_file
	(\ioerror -> appendChan stdout
				("Can't open " ++ gram_file)
				abort done)
	(\grammar_text ->
    let { (parser,error_messages)
		= parse_gen' grammar_text main_name scan_name }
    in
    appendChan stdout error_messages abort
    (writeFile parser_file parser
	(\ioerror -> appendChan stdout
				("Can't write to " ++ parser_file)
				abort done)
	done))

parse_gen' grammar_text main_name scan_name
	= (((make_datatype (nonterminals gram) ++) .
	   ("\n" ++) .
	   ("data Loop a b = Loop (a -> [Loop a b]-> b)\n\n" ++) .
	   (make_nonterminal_numbers (nonterminals gram) ++) .
	   ("\n" ++) .
	   ((main_name ++
              " text = get_SLR_val (parse_state1 [Loop go_to_1] [] (" ++
	     scan_name ++ " text))\n\n") ++) .
	   (get_SLR_val_text ++) .
	   unlines' (map make_parse_state table) .
	   ("\n" ++) .
	   unlines' (map make_reduce
		       (zip3 [1..] (productions gram) sem_funs)) .
	   ("\n" ++) .
	   unlines' (map make_gotos table) .
	   ("\n" ++ ))
	  auxils,
	  error_messages)

	where
	  (gram, sem_funs, auxils)
		= parse_gram grammar_text

	  table = slr_table gram

	  error_messages = check_conflicts gram table

	  make_datatype nonts
		= "data SLR_val " ++
		  unwords (map ("type_"++) nonts) ++
		  " type_terminal inputType\n  = Terminal_val_ type_terminal\n" ++
		  "  | Parse_Error Int inputType\n" ++
		  unlines (map (\n->"  | " ++ n ++ "_val type_" ++ n) nonts)

	  make_nonterminal_numbers nonts
		= unlines (map (\(num,nont) ->
				"num_" ++ nont ++ " = " ++ show num)
			       (zip [1..] nonts))

	  get_SLR_val_text
		= "get_SLR_val val\n  = case val of {\n" ++
		  "     " ++ head (nonterminals gram) ++
			"_val root_val -> root_val;\n" ++
		  "     Parse_Error state input -> parse_error state input}\n\n"

	  make_reduce (prod,(nont,rhs),sem_fun)
		= "reduce_" ++ show prod ++
	   	  " (" ++ take (2*length rhs) (cycle "_:")
	   	  ++ "states@(Loop state : _))\n"
	   	  ++ "          (" ++
	   	  concat (map (\(sym,num) ->
			  (make_con sym) ++ " x"++ (show num) ++ " : ")
		   (reverse (zip rhs [1..])))
	   	  ++ "rest)\n"
	   	  ++ "    = state num_" ++ nont ++ " states (("
	   	  ++ nont ++"_val (" ++ sem_fun ++")) : rest)\n"

	  make_con (Left t)  = "Terminal_val_"
	  make_con (Right n) = n ++"_val"

	  make_parse_state (state,actions,red_end,gotos)
		= "parse_state" ++ show state ++ " states stack []\n" ++
		  "  = " ++ make_red_end state red_end ++ "\n" ++
		  "parse_state" ++ show state ++
		  " states stack" ++ 
		  " input@((s_t,s_v,_):rest)\n" ++
		  "  = " ++ make_do_actions state
				(zip (terminals gram) actions)
		  ++ "\n"

	  make_red_end state []
		= "Parse_Error " ++ show state ++ " []\n"
	  make_red_end state (0:_)
		= "head stack\n"
	  make_red_end state [r1]
		= "reduce_" ++ show r1 ++" states stack []\n"
	  make_red_end state (r1:rs)
		= "case reduce_" ++ show r1 ++" states stack [] of {\n"
		  ++ "    Parse_Error _ _ -> " ++ make_red_end state rs
		  ++ ";\n    val -> val}\n"

	  make_do_actions state actions
		= "case s_t of\n" ++
		  concat (map (make_do_action state) actions) ++
		  "     _ -> Parse_Error " ++ show state ++ " input\n"

	  make_do_action state (t,(0,[]))
		= ""
	  make_do_action state (t,(shift,(0:rs)))
		= make_do_action state (t,(shift,rs))
	  make_do_action state (t,(0,rs))
		= "     tok | tok == " ++ t ++ " -> " ++
		  make_reduce_actions state rs
	  make_do_action state (t,(shift,[]))
		= "     tok | tok == " ++ t ++ " -> " ++
		  "parse_state" ++ show shift ++ " (Loop go_to_" ++
		   show shift ++
		  " : states) ((Terminal_val_ s_v):stack) rest\n"
	  make_do_action state (t,(shift,rs))
		= "     tok | tok == " ++ t ++ " -> " ++
		  "case parse_state" ++ show shift ++ " (Loop go_to_" ++
		   show shift ++
		  " : states) ((Terminal_val_ s_v):stack) rest of {\n"
		  ++ "    Parse_Error _ _ -> " ++ make_reduce_actions state rs
		  ++ ";\n    val -> val}\n"

	  make_reduce_actions state [r1]
		= "reduce_" ++ show r1 ++" states stack input\n"
	  make_reduce_actions state (r1:rs)
		= "case reduce_" ++ show r1 ++" states stack input of {\n"
		  ++ "    Parse_Error _ _ -> " ++ make_reduce_actions state rs
		  ++ ";\n    val -> val}\n"

	  make_gotos (state,actions,red_end,gotos)
	    | all (0 ==) gotos
		= "go_to_" ++ show state ++
		  " = go_to_" ++ show state ++ "\n"
	  make_gotos (state,actions,red_end,gotos)
	    | otherwise
		= "go_to_" ++ show state ++ " nont states\n"++
		  concat (map make_goto
			      (zip (nonterminals gram) gotos))

	  make_goto (nont,0)
		= ""
	  make_goto (nont,go_state)
		= "   | nont == num_" ++ nont
		  ++ " = parse_state" ++ show go_state ++
		  " (Loop go_to_" ++ show go_state ++
		  " : states)\n"

	  unlines' []    = id
	  unlines' (a:l) = (a ++) . ('\n' :) . unlines' l

-- pretty-print functions for grammars, collections etc.

show_grammar' gram
	= unlines (map show_prod (zip [1..] (productions gram)))
	where
	  show_prod (p,(n,rhs))
		= rjustify 2 (show p) ++ " " ++
		  rjustify 12 (take 12 n) ++ " ->" ++
		   concat (map show_either rhs)
	  show_either (Left t)  = ' ':t
	  show_either (Right n) = ' ':n

show_collection' gram
	= (layn . (map (unlines . ("":) . reverse . (map show_item))))
	    (collection_of_item_sets gram (goal_symbol gram ++ "\'") \\ [[]])
	where
	  show_item (n,before,after)
		= rjustify 12 (take 12 n) ++ " ->" ++
		  concat (map show_either (reverse before))
		  ++ " ." ++
		  concat (map show_either after)
	  show_either (Left t)  = ' ':t
	  show_either (Right n) = ' ':n

show_table' gram
	= show_grammar' gram ++ "\n"++
	  "    |" ++
	  concat (map (cjustify 8 . take 8) (terminals gram)) ++
	  cjustify 8 "END" ++ "|" ++
	  concat (map (rjustify 5 . take 4) (nonterminals gram)) ++
	  "\n" ++
	  take (14 + 8 * length (terminals gram) +
		5 * length (nonterminals gram))
	       (repeat '-') ++ "\n" ++
	  unlines (map show_row table) ++ "\n" ++
	  check_conflicts gram table
	where
	  table = slr_table gram
	  prod_to_num = zip (productions gram) [1..]
	  show_row (state,actions,red_end,gotos)
		= rjustify 3 (show state) ++ " |" ++
		  concat (map show_action actions) ++
		  cjustify 8 (take 8 (concat (map show_red red_end))) ++
		  "|" ++ concat (map (rjustify 5 . take 5 . show_state) gotos)
	  show_action (shift,reduce)
		= cjustify 8 (take 8 (show_shift shift ++
				concat (map show_red reduce)))
	  show_shift 0 = ""
	  show_shift s  = "s"++ show s
	  show_red 0 = "acc"
	  show_red p = "r" ++ show p
	  show_state 0 = ""
	  show_state s  = show s

readfile file s2s
  = readFile file
	(\ioerror -> appendChan stdout
				("Can't open " ++ file)
				abort done)
	((\st -> appendChan stdout st abort done) . s2s)

show_grammar gram_file
	= readfile gram_file (show_grammar'. fst3 . parse_gram)

show_collection gram_file
	= readfile gram_file (show_collection' . fst3 . parse_gram)

show_table gram_file
	= readfile gram_file (show_table' . fst3 . parse_gram)

main = readChan stdin abort (\inp -> show_table (head (lines inp)))
